        page    ,132
;*****************************************************************************
;  Terminal Program for dummy and intelligent Controller Device Driver
;*****************************************************************************

critical_err_int  equ     24h         ; DOS critical error interrupt number

normal_attr       equ     7           ; normal video attribute

binary            equ     0020h       ; binary I/O mode mask

; file/device access modes
read_only         equ     0
write_only        equ     1
read_write        equ     2

; errors
dderra            equ     0ah         ; write fault
dderrb            equ     0bh         ; read fault
dderrc            equ     0ch         ; general failure

stack   segment stack 'stack'
       dw      64 dup ('@@')
stack   ends

data    segment public 'data'
int24h_offset     dw      ?           ; critical error handler interrupt vector
int24h_segment    dw      ?

device_err_code   db      ?           ; device driver error code

line_buffer       db      9           ; 8 character device name + CR
                  db      ?           ; returned count
port_name         db      9 dup (?)   ; keyboard buffer

handle            dw      ?           ; file handle

read_buffer       db      8  dup (?)  ; read data buffer
write_buffer      db      8  dup (?)  ; write data buffer
ioctl_buffer      db      16 dup (?)  ; I/O control buffer

device_msg        db      'enter device name: $'

logo_msg          db      'Terminal Program  (enter F10 to exit)$'

open_err_msg      db      0dh,0ah,'open error',07h,0dh,0ah,'$'
close_err_msg     db      0dh,0ah,'close error',07h,0dh,0ah,'$'

tbof_error        equ     16          ; transmit buffer overflow error #

error_tbl         equ     $
      dw      err_0001
      dw      err_0002
      dw      err_0004
      dw      err_0008
      dw      err_0010
      dw      err_0020
      dw      err_0040
      dw      err_0080
      dw      err_0100
      dw      err_0200
      dw      err_0400
      dw      err_0800
      dw      err_1000
      dw      err_2000
      dw      err_4000
      dw      err_8000

      dw      err_tbof

err_0001        db      '0001: Undefined Error',07h,'$'
err_0002        db      'Overrun Error',07h,'$'
err_0004        db      'Parity Error',07h,'$'
err_0008        db      'Framing Error',07h,'$'
err_0010        db      'Break Interrupt',07h,'$'
err_0020        db      '0020: Undefined Error',07h,'$'
err_0040        db      '0040: Undefined Error',07h,'$'
err_0080        db      '0080: Undefined Error',07h,'$'
err_0100        db      'Receive Buffer Overrun',07h,'$'
err_0200        db      '0200: Undefined Error',07h,'$'
err_0400        db      '0400: Undefined Error',07h,'$'
err_0800        db      '0800: Undefined Error',07h,'$'
err_1000        db      'CTS Inactive',07h,'$'
err_2000        db      'DSR Inactive',07h,'$'
err_4000        db      'RI Transitioned',07h,'$'
err_8000        db      'DCD Inactive',07h,'$'

err_tbof        db      'Transmit Buffer Full',07h,'$'

clear_msg       db      '                      $'
data    ends

code    segment public 'code'
        assume  cs:code, ds:data

;  name: main
;  function: perform initialization and termination processing
;
;  enter: from DOS w/ DS=> Program Segment Prefix
;  exit: to DOS

main    proc    far
        push    ds                      ; save DOS return
        xor     ax,ax
        push    ax
        mov     ax,data                 ; establish data segment
        mov     ds,ax

        mov     al,24h                  ; AL = interrupt #
        mov     ah,35h                  ; AH = function = get vector
        int     21h
        mov     int24h_offset,bx        ; save current critical error
        mov     int24h_segment,es       ;  interrupt vector

        mov     dx,offset critical_error; DS:DX=> new critical error
        mov     ax,cs                   ;  interrupt vector
        mov     ds,ax
        mov     al,24h                  ; AL = interrupt #
        mov     ah,25h                  ; AH= function = set interrupt vector
        int     21h

        mov     ax,data                 ; re-establish data segment
        mov     ds,ax

        call    get_device              ; get device name from user
; return w/ DS:DX=> device name

        mov     al,read_write           ; AL = access code
        mov     ah,3dh                  ; AH = function = open a file
        int     21h
        jnc     main_10                 ; continue if successful...
        mov     dx,offset open_err_msg  ; otherwise display error message
        mov     ah,9
        int     21h
        jmp     main_99                 ; exit ...
main_10:
        mov     handle,ax               ; save file handle
; switch I/O mode to binary ...
        mov     bx,ax                   ; BX = file handle
        mov     al,0                    ; AL = function (get device info)
        mov     ah,44h                  ; AH = function = IOCTL
        int     21h
        or      dx,binary               ; indicate binary mode
        xor     dh,dh                   ;  and clear upper byte
        mov     al,1                    ; AL = function (set device info)
        mov     ah,44h                  ; AH = function = IOCTL
        int     21h
; flush transmit and receive buffers ...
        mov     al,3                    ; AL = function (write IOCTL string)
        mov     write_buffer,1          ; setup control string
        mov     dx,offset write_buffer  ;  to flush transmit buffer
        mov     ah,44h                  ; AH = function = IOCTL
        mov     cx,1
        int     21h
        mov     al,3                    ; AL = function (write IOCTL string)
        mov     write_buffer,2          ; setup control string
        mov     dx,offset write_buffer  ;  to flush receive buffer
        mov     ah,44h                  ; AH = function = IOCTL
        mov     cx,1
        int     21h

;       mov     al,3                    ; AL = function (write IOCTL string)
;       mov     byte ptr write_buffer,3 ; setup control string
;       mov     byte ptr write_buffer+1,1
;       mov     word ptr write_buffer+2,4800 ; baud rate
;       mov     dx,offset write_buffer  ;  to flush receive buffer
;       mov     ah,44h                  ; AH = function = IOCTL
;       mov     cx,3
;       int     21h

        call    init_display            ; initialize the display
        call    terminal                ; do terminal processing
        call    clr_home                ; clear the screen/home cursor

        mov     bx,handle               ; BX = file handle
        mov     ah,3eh                  ; AH= function = close a file handle
        int     21h
        jnc     main_20                 ; continue if successful...
        mov     dx,offset close_err_msg ; otherwise display error message
        mov     ah,9
        int     21h                     ; exit ...
main_20:
        lds     dx,dword ptr int24h_offset  ; DS:DX=> original critical
                                            ;  error interrupt vector
        mov     al,24h                  ; AL = interrupt #
        mov     ah,25h                  ; AH= function = set interrupt vector
        int     21h
main_99:
        ret
main    endp

;  name: get_device
;  function: query user for device name
;
;  enter: no parameters required
;  exit: 'line_buffer' contains device name
;
;  assume all registers are altered

get_device      proc
        call    clr_home                ; clear the screen/home cursor
        mov     dx,offset device_msg    ; DS:DX=> query message
        mov     ah,9                    ; solicit device name
        int     21h
        mov     dx,offset line_buffer   ; DS:DX=> input buffer
        mov     ah,0ah                  ; get buffered input
        int     21h
        mov     dx,offset port_name     ; DX=> device name
        mov     bx,dx                   ; BX=> device name
        mov     al,line_buffer+1        ; get device name length
        xor     ah,ah                   ; clear upper byte
        add     bx,ax                   ; offset to end of device name+1
        mov     [bx],ah                 ; terminate name w/ a null
        ret
get_device      endp

;  name: terminal
;  function: terminal processing
;
;  enter: no parameters required
;  exit: no parameters returned
;
;  assume all registers are altered

terminal        proc
        call    device_error            ; poll the error channel
        call    receive_character       ; check receive buffer
        or      ax,ax                   ; character returned ?
        jz      terminal_10             ;  NO ... continue
        mov     al,read_buffer          ; get character
        call    display_character       ; display it
terminal_10:
        mov     dx,0ffh                 ; DX = 0ffh, for console input
        mov     ah,06h                  ; AH= function = direct console I/O
        int     21h
        jz      terminal                ; continue if no character...
        or      al,al                   ; extended code ?
        jz      terminal_12             ;  YES ... handle it
        call    transmit_character      ; transmit character
        jmp     terminal                ; continue...
; get extended code and check for program exit ...
terminal_12:
        mov     ah,06h                  ; AH= function = direct console I/O
        int     21h
        cmp     al,44h                  ; function key F-10 ?
        jnz     terminal                ;  NO ... continue
        ret                             ; otherwise exit the program ...
terminal        endp

;  name: receive_character
;  function: remove character from receive buffer
;
;  enter: no parameters required
;  exit: AX = 1 if a character was received, AX = 0 if receive buffer
;        was empty or an error occurred
;
;  assume all registers are altered

receive_character       proc
        mov     device_err_code,0       ; clear device error code
        mov     dx,offset read_buffer   ; DS:DX=> buffer
        mov     cx,1                    ; CX = number of bytes to be read
        mov     bx,handle               ; BX = file handle
        mov     ah,3fh                  ; AH= function = read from device
        int     21h
        jnc     receive_character_99    ; exit if no error...
        mov     al,device_err_code      ; get device error code
        call    device_error            ; process error
        xor     al,al                   ; indicate no characters returned
receive_character_99:
        ret
receive_character       endp

;  name: transmit_character
;  function: add character to transmit buffer
;
;  enter: AL = character
;  exit: no parameters returned; a transmit buffer full condition will
;        cause characters to be lost
;
;  assume all registers are altered

transmit_character      proc
        call    display_character       ; display character
        mov     write_buffer,al         ; move character to buffer
transmit_character_02:
        mov     device_err_code,0       ; clear device error code
        mov     dx,offset write_buffer  ; DS:DX=> buffer
        mov     cx,1                    ; CX = number of bytes to write
        mov     bx,handle               ; BX = file handle
        mov     ah,40h                  ; AH = function = write to device
        int     21h
        jnc     transmit_character_99   ; exit if no error...
        mov     al,device_err_code      ; get device error code
        cmp     al,dderra               ; write fault (transmit buffer full)?
        jnz     transmit_character_10   ;  NO ... continue
        mov     cl,tbof_error           ; setup error #
        call    display_error           ;  and display it
        jmp     short transmit_character_99     ; exit ...
transmit_character_10:
        call    device_error            ; process other errors...
transmit_character_99:
        ret
transmit_character      endp

;  name: init_display
;  function: initialize display for terminal processing
;
;  enter: no parameters required
;  exit: no parameters returned
;
;  assume all registers are altered

init_display    proc
        call    clr_home                ; clear the screen/home cursor
; display the 25th line logo
        mov     bh,0                    ; BH = display page
        mov     dx,1800h                ; DX = cursor position (row/column)
        mov     ah,2                    ; set cursor position
        int     10h                     ;  via ROM BIOS routine
        mov     si,offset logo_msg      ; DS:BX=> logo message
        call    display_message         ; display it
; home the cursor
init_display_20:
        mov     dx,0                    ; DX = cursor position (row/column)
        mov     ah,2                    ; set cursor position
        int     10h                     ;  via ROM BIOS routine
        ret
init_display    endp

;  name: clr_home
;  function: clear the screen and position cursor to home location
;
;  enter: no parameters required
;  exit: no parameters returned
;
;  assume all registers are altered

clr_home        proc
; position cursor to home location
        mov     bh,0                    ; BH = display page
        mov     dx,0                    ; DX = cursor position (row/column)
        mov     ah,2                    ; position the cursor
        int     10h                     ;  via ROM BIOS routine
; clear the screen
        mov     bl,7                    ; BL = attribute of character
        mov     cx,25*80                ; CX = count of characters to write
        mov     al,' '                  ; AL = character to write
        mov     ah,10                   ; clear screen
        int     10h                     ;  via ROM BIOS routine
        ret
clr_home        endp

;  name: display_character
;  function: the character will be displayed, the cursor will be advanced to
;            the next location and the screen will be scrolled as required; no
;            characters will be written on the 25th line
;
;  enter: AL = character
;  exit: AL = character
;
;  registers BH, CX and DX altered

display_character       proc
        push    ax                      ; save character
        mov     bh,0                    ; BH = display page
        mov     cx,1                    ; CX = character count
        mov     ah,14                   ; display character
        int     10h                     ;  via ROM BIOS routine

        mov     ah,3                    ; read cursor position
        int     10h                     ;  via ROM BIOS routine

        cmp     dh,24                   ; end of display area?
        jnz     display_character_10    ;  NO ... exit
        push    dx                      ; save cursor position
        mov     al,1                ; AL= # of blank lines at bottom of window
        mov     bh,7                ; BH= attribute to be used on blank line
        mov     cx,0                ; CX= row/column of upper left corner of the
                                    ;     scroll area
        mov     dx,174fh            ; DX= row/column of lower right corner of
                                    ;     the scroll area
        mov     ah,6                    ; scroll active page up
        int     10h                     ;  via ROM BIOS routine

        mov     bh,0                    ; BH = display page
        pop     dx                      ; recover cursor position
        mov     bh,0                    ; BH = display page
        mov     dh,23                   ; reset row #
        mov     ah,2                    ; update cursor position
        int     10h                     ;  via ROM BIOS routine
display_character_10:
        pop     ax                      ; recover character
        ret
display_character       endp

;  name: device_error
;  function: poll error channel and process any errors
;
;  enter: no parameters required
;  exit: no parameters returned
;
;  assume all registers are altered

device_error    proc
        call    ioc_input               ; poll error channel
        jz      device_error_99         ; exit if no errors...
        mov     cx,0                    ; initialize loop count/error #
device_error_02:
        shr     ax,1                    ; shift low bit into carry
        jnc     device_error_10         ; continue if no error...
        call    display_error           ; display error message
device_error_10:
        inc     cx                      ; bump loop count/error #
        cmp     cx,16                   ; finished?
        jc      device_error_02         ;  NO ... continue
device_error_99:
        ret
device_error    endp

;  name: ioc_input
;  function: poll error channel
;
;  enter: no parameters required
;  exit: zero flag clear if an error has occurred and AX = error flags,
;        else zero flag set if no errors have occurred
;
;  assume all registers are altered

ioc_input       proc
        mov     dx,offset ioctl_buffer  ; DS:DX=> I/O control buffer
        mov     cx,07h                  ; CX = # of bytes to read
        mov     bx,handle               ; BX = file handle
        mov     al,2                    ; AL= function = read ioctl data
        mov     ah,44h                  ; AH= function = IOCTL
        int     21h
        mov     ax,word ptr ioctl_buffer; get error flags
        or      ax,ax                   ; test flags for return
        ret
ioc_input       endp

;  name: display_error
;  function: display error message on 25th line
;
;  enter: AL = error message #
;  exit: no parameters returned
;
;  register SI altered

display_error   proc
        push    ax                      ; save registers
        push    bx
        push    cx
        push    dx
        push    cx                      ; save error code
        mov     bh,0                    ; BH = display page
        mov     ah,3                    ; read cursor position
        int     10h                     ;  via ROM BIOS routine

        mov     cx,dx                   ; save cursor position
        mov     dx,1833h                ; DX = cursor position (row/column)
        mov     ah,2                    ; position cursor
        int     10h                     ;  via ROM BIOS routine

        pop     ax                      ; recover error code
        push    cx                      ;  and save cursor position
        mov     bx,offset error_tbl     ; BX=> error message table
        xor     ah,ah                   ; clear upper byte of error code
        shl     ax,1                    ;  and double it
        add     bx,ax                   ; offset into table w/ error code
        mov     si,[bx]                 ; setup pointer to error message
        call    display_message         ;  and display it
        xor     cx,cx                   ; delay before returning control
display_error_10:
        loop    display_error_10

        mov     dx,1833h                ; DX = cursor position (row/column)
        mov     ah,2                    ; position cursor
        int     10h                     ;  via ROM BIOS routine

        mov     si,offset clear_msg     ; setup pointer to 'clear' message
        call    display_message         ;  and display it
        pop     dx                      ; recover cursor position
        mov     bh,0                    ; BH = display page
        mov     ah,2                    ; restore cursor position
        int     10h                     ;  via ROM BIOS routine
        pop     dx                      ; recover registers
        pop     cx
        pop     bx
        pop     ax
        ret
display_error   endp

;  name: display_message
;  function: display message string on video; the cursor must have been
;            previously positioned
;
;  enter: DS:SI=> message string terminated by '$'
;  exit: DS:SI=> string terminator ('$')
;
;  registers AX, BH, CX and SI altered

display_message proc
        mov     bh,0                    ; BH = display page
        mov     cx,1                    ; CX = count of characters to write
display_message_10:
        lodsb                           ; AL = character
        cmp     al,'$'                  ; end of string?
        jz      display_message_99      ;  YES ... finished
        mov     ah,14                   ; display character
        int     10h                     ;  via ROM BIOS routine
        jmp     display_message_10      ; continue...
display_message_99:
        ret
display_message endp

;  name: critical_error
;  function: this procedure replaces the DOS int 24h critical error handler
;
;  enter: BP:SI=> Device Header Control Block
;            DI = device error code
;  exit: AL = 3 (informs DOS to fail the system call that is in progress)
;
;  no other registers altered

critical_error  proc
        push    ds                      ; save the DOS data segment
        mov     ax,data                 ;  and setup local data segment
        mov     ds,ax
        mov     ax,di                   ; get error code
        mov     device_err_code,al      ;  and save it locally
        mov     al,3                    ; fail the system call in progress
        pop     ds                      ; recover data segment
        iret
critical_error  endp

code    ends
        end     main

